@@ -1,6 +1,6 @@
 /* gpsutils.c -- code shared between low-level and high-level interfaces
  *
- * This file is Copyright 2010 by the GPSD project
+ * This file is Copyright by the GPSD project
  * SPDX-License-Identifier: BSD-2-clause
  */
 
@@ -9,7 +9,7 @@
  * newer features (like clock_gettime).  See the POSIX spec for more info:
  * http://pubs.opengroup.org/onlinepubs/9699919799/functions/V2_chap02.html#tag_15_02_01_02 */
 
-#include "../include/gpsd_config.h"  /* must be before all includes */
+#include "../include/gpsd_config.h"    // must be before all includes
 
 #include <ctype.h>
 #include <errno.h>
@@ -18,8 +18,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <sys/select.h>  /* for to have a pselect(2) prototype a la POSIX */
-#include <sys/time.h>    /* for to have a pselect(2) prototype a la SuS */
+#include <sys/select.h>  // for to have a pselect(2) prototype a la POSIX
+#include <sys/time.h>    // for to have a pselect(2) prototype a la SuS
 #include <time.h>
 
 #include "../include/gps.h"
@@ -32,6 +32,99 @@
 #include <QStringList>
 #endif
 
+// gnssId names
+const struct vlist_t vgnssId[] = {
+    {0, "GPS"},
+    {1, "SBAS"},
+    {2, "GAL"},
+    {3, "BDS"},
+    {4, "IMES"},
+    {5, "QZSS"},
+    {6, "GLO"},
+    {7, "NavIC"},
+    {0, NULL},
+};
+
+// mode val to mode string
+const struct vlist_t vmode_str[] = {
+    {1, "No Fix"},
+    {2, "2D Fix"},
+    {3, "3D Fix"},
+    {0, NULL},
+};
+
+// status val to status string
+const struct vlist_t vstatus_str[] = {
+    {0, "UNK"},
+    {1, "GPS"},
+    {2, "DGPS"},
+    {3, "RTK_FIX"},
+    {4, "RTK_FLT"},
+    {5, "DR"},
+    {6, "GNSSDR"},
+    {7, "TIME"},      // Surveyd
+    {8, "SIM "},
+    {0, NULL},
+};
+
+/* char2str(ch, vlist) - given a char, return a matching string.
+ *
+ * Return: pointer to string
+ */
+const char *char2str(unsigned char ch, const struct clist_t *clist)
+{
+    while (NULL != clist->str) {
+        if (clist->ch == ch) {
+            return clist->str;
+        }
+        clist++;
+    }
+    return "Unk";
+}
+
+/* flags2str(val, vlist) - given flags, return a matching string.
+ *
+ * flags the flags to find in vlist
+ * buffer -  the buffer to return string in
+ ( buflen - length of buffer
+ *
+ * Return: pointer to passed in buffer
+ *         A string matching the flags.
+ */
+const char *flags2str(unsigned long flags, const struct flist_t *flist,
+                      char *buffer, size_t buflen)
+{
+    buffer[0] = '\0';
+    while (NULL != flist->str) {
+        if (flist->val == (flist->mask & flags)) {
+            if ('\0' != buffer[0]) {
+                strlcat(buffer, ",", buflen);
+            }
+            strlcat(buffer, flist->str, buflen);
+        }
+        flist++;
+    }
+    return buffer;
+}
+
+/* val2str(val, vlist) - given a value, return a matching string.
+ *
+ * val the value to find in vlist
+ *
+ * Return: pointer to static string
+ *         The string matching val, or "Unk".
+ */
+const char *val2str(unsigned long val, const struct vlist_t *vlist)
+{
+    while (NULL != vlist->str) {
+        if (vlist->val == val) {
+            return vlist->str;
+        }
+        vlist++;
+    }
+    return "Unk";
+}
+
 /*
  * Berkeley implementation of strtod(), inlined to avoid locale problems
  * with the decimal point and stripped down to an atof()-equivalent.
@@ -78,7 +171,7 @@ double safe_atof(const char *string)
     double fraction, dblExp, *d;
     const char *p;
     int c;
-    int exp = 0;                /* Exponent read from "EX" field. */
+    int exp = 0;                // Exponent read from "EX" field.
     int fracExp = 0;            /* Exponent that derives from the fractional
                                  * part.  Under normal circumstatnces, it is
                                  * the negative of the number of digits in F.
@@ -88,7 +181,7 @@ double safe_atof(const char *string)
                                  * unnecessary overflow on I alone).  In this
                                  * case, fracExp is incremented one for each
                                  * dropped digit. */
-    int mantSize;               /* Number of digits in mantissa. */
+    int mantSize;               // Number of digits in mantissa.
     int decPt;                  /* Number of mantissa digits BEFORE decimal
                                  * point. */
     const char *pExp;           /* Temporarily holds location of exponent
@@ -144,7 +237,7 @@ double safe_atof(const char *string)
     if (decPt < 0) {
         decPt = mantSize;
     } else {
-        mantSize -= 1;                  /* One of the digits was the point. */
+        mantSize -= 1;                  // One of the digits was the point.
     }
     if (mantSize > 18) {
         fracExp = decPt - 18;
@@ -154,27 +247,26 @@ double safe_atof(const char *string)
     }
     if (mantSize == 0) {
         fraction = 0.0;
-        //p = string;
+        // p = string;
         goto done;
     } else {
         int frac1, frac2;
+
         frac1 = 0;
-        for ( ; mantSize > 9; mantSize -= 1)
-        {
+        for ( ; mantSize > 9; mantSize -= 1) {
             c = *p;
             p += 1;
-            if (c == '.') {
+            if ('.' == c) {
                 c = *p;
                 p += 1;
             }
             frac1 = 10*frac1 + (c - '0');
         }
         frac2 = 0;
-        for (; mantSize > 0; mantSize -= 1)
-        {
+        for (; mantSize > 0; mantSize -= 1) {
             c = *p;
             p += 1;
-            if (c == '.') {
+            if ('.' == c) {
                 c = *p;
                 p += 1;
             }
@@ -188,19 +280,27 @@ double safe_atof(const char *string)
      */
 
     p = pExp;
-    if ((*p == 'E') || (*p == 'e')) {
+    if (('E' == *p) ||
+        ('e' == *p)) {
         p += 1;
-        if (*p == '-') {
+        if ('-' == *p) {
             expSign = true;
             p += 1;
         } else {
-            if (*p == '+') {
+            if ('+' == *p) {
                 p += 1;
             }
             expSign = false;
         }
         while (isdigit((int) *p)) {
             exp = exp * 10 + (*p - '0');
+            if (1024 < exp) {
+                if (true == expSign) {
+                    // exponent underflow!
+                    return 0.0;
+                } // else  exponent overflow!
+                return INFINITY;
+            }
             p += 1;
         }
     }
@@ -217,7 +317,7 @@ double safe_atof(const char *string)
      * fraction.
      */
 
-    if (exp < 0) {
+    if (0 > exp) {
         expSign = true;
         exp = -exp;
     } else {
@@ -246,9 +346,21 @@ done:
     return fraction;
 }
 
-#define MONTHSPERYEAR   12      /* months per calendar year */
+#define MONTHSPERYEAR   12      // months per calendar year
 
-/* stuff a fix structure with recognizable out-of-band values */
+// clear a baseline_t
+static void gps_clear_base(struct baseline_t *base)
+{
+    base->status = STATUS_UNK;
+    base->east = NAN;
+    base->north = NAN;
+    base->up = NAN;
+    base->length = NAN;
+    base->course = NAN;
+    base->ratio = NAN;
+}
+
+// stuff a fix structure with recognizable out-of-band values
 void gps_clear_fix(struct gps_fix_t *fixp)
 {
     memset(fixp, 0, sizeof(struct gps_fix_t));
@@ -273,7 +385,7 @@ void gps_clear_fix(struct gps_fix_t *fixp)
     fixp->sep = NAN;
     fixp->speed = NAN;
     fixp->track = NAN;
-    /* clear ECEF too */
+    // clear ECEF too
     fixp->ecef.x = NAN;
     fixp->ecef.y = NAN;
     fixp->ecef.z = NAN;
@@ -291,42 +403,68 @@ void gps_clear_fix(struct gps_fix_t *fixp)
     fixp->geoid_sep = NAN;
     fixp->dgps_age = NAN;
     fixp->dgps_station = -1;
+    fixp->temp = NAN;
     fixp->wanglem = NAN;
     fixp->wangler = NAN;
     fixp->wanglet = NAN;
     fixp->wspeedr = NAN;
     fixp->wspeedt = NAN;
+    fixp->wtemp = NAN;
+    gps_clear_base(&fixp->base);
 }
 
-/* stuff an attitude structure with recognizable out-of-band values */
+// stuff an attitude structure with recognizable out-of-band values
 void gps_clear_att(struct attitude_t *attp)
 {
     memset(attp, 0, sizeof(struct attitude_t));
-    attp->pitch = NAN;
-    attp->roll = NAN;
-    attp->yaw = NAN;
-    attp->dip = NAN;
-    attp->mag_len = NAN;
-    attp->mag_x = NAN;
-    attp->mag_y = NAN;
-    attp->mag_z = NAN;
     attp->acc_len = NAN;
     attp->acc_x = NAN;
     attp->acc_y = NAN;
     attp->acc_z = NAN;
+    attp->depth = NAN;
+    attp->dip = NAN;
+    attp->gyro_temp = NAN;
     attp->gyro_x = NAN;
     attp->gyro_y = NAN;
+    attp->gyro_z = NAN;
+    attp->heading = NAN;
+    attp->mheading = NAN;
+    attp->mag_len = NAN;
+    attp->mag_x = NAN;
+    attp->mag_y = NAN;
+    attp->mag_z = NAN;
+    attp->pitch = NAN;
+    attp->roll = NAN;
+    attp->rot = NAN;
     attp->temp = NAN;
-    attp->depth = NAN;
+    attp->yaw = NAN;
+    gps_clear_base(&attp->base);
 }
 
+// Clear a dop_t structure
 void gps_clear_dop( struct dop_t *dop)
 {
     dop->xdop = dop->ydop = dop->vdop = dop->tdop = dop->hdop = dop->pdop =
         dop->gdop = NAN;
 }
 
-/* stuff a log structure with recognizable out-of-band values */
+// Clear a gst structure
+void gps_clear_gst( struct gst_t *gst)
+{
+    memset(&gst->utctime, 0, sizeof(gst->utctime));
+    gst-> rms_deviation =  NAN;
+    gst-> smajor_deviation =  NAN;
+    gst-> sminor_deviation =  NAN;
+    gst-> smajor_orientation =  NAN;
+    gst-> lat_err_deviation =  NAN;
+    gst-> lon_err_deviation =  NAN;
+    gst-> alt_err_deviation =  NAN;
+    gst-> ve_err_deviation =  NAN;
+    gst-> vn_err_deviation =  NAN;
+    gst-> vu_err_deviation =  NAN;
+}
+
+// stuff a log structure with recognizable out-of-band values
 void gps_clear_log(struct gps_log_t *logp)
 {
     memset(logp, 0, sizeof(struct gps_log_t));
@@ -357,16 +495,19 @@ void gps_merge_fix(struct gps_fix_t *to,
                    gps_mask_t transfer,
                    struct gps_fix_t *from)
 {
-    if ((NULL == to) || (NULL == from))
+    if ((NULL == to) ||
+        (NULL == from)) {
         return;
-    if ((transfer & TIME_SET) != 0)
+    }
+    if (0 != (transfer & TIME_SET)) {
         to->time = from->time;
-    if ((transfer & LATLON_SET) != 0) {
+    }
+    if (0 != (transfer & LATLON_SET)) {
         to->latitude = from->latitude;
         to->longitude = from->longitude;
     }
     if (0 != (transfer & MODE_SET)) {
-        /* FIXME?  Maybe only upgrade mode, not downgrade it */
+        // FIXME?  Maybe only upgrade mode, not downgrade it
         to->mode = from->mode;
     }
     /* Some messages only report mode, some mode and status, some only status.
@@ -387,9 +528,10 @@ void gps_merge_fix(struct gps_fix_t *to,
             to->depth = from->depth;
         }
     }
-    if ((transfer & TRACK_SET) != 0)
+    if (0 != (transfer & TRACK_SET)) {
         to->track = from->track;
-    if ((transfer & MAGNETIC_TRACK_SET) != 0) {
+    }
+    if (0 != (transfer & MAGNETIC_TRACK_SET)) {
         if (0 != isfinite(from->magnetic_track)) {
             to->magnetic_track = from->magnetic_track;
         }
@@ -397,12 +539,15 @@ void gps_merge_fix(struct gps_fix_t *to,
             to->magnetic_var = from->magnetic_var;
         }
     }
-    if ((transfer & SPEED_SET) != 0)
+    if (0 != (transfer & SPEED_SET)) {
         to->speed = from->speed;
-    if ((transfer & CLIMB_SET) != 0)
+    }
+    if (0 != (transfer & CLIMB_SET)) {
         to->climb = from->climb;
-    if ((transfer & TIMERR_SET) != 0)
+    }
+    if (0 != (transfer & TIMERR_SET)) {
         to->ept = from->ept;
+    }
     if (0 != isfinite(from->epx) &&
         0 != isfinite(from->epy)) {
         to->epx = from->epx;
@@ -417,26 +562,27 @@ void gps_merge_fix(struct gps_fix_t *to,
     if (0 != isfinite(from->eps)) {
         to->eps = from->eps;
     }
-    /* spherical error probability, not geoid separation */
+    // spherical error probability, not geoid separation
     if (0 != isfinite(from->sep)) {
         to->sep = from->sep;
     }
-    /* geoid separation, not spherical error probability */
+    // geoid separation, not spherical error probability
     if (0 != isfinite(from->geoid_sep)) {
         to->geoid_sep = from->geoid_sep;
     }
     if (0 != isfinite(from->epv)) {
         to->epv = from->epv;
     }
-    if ((transfer & SPEEDERR_SET) != 0)
+    if (0 != (transfer & SPEEDERR_SET)) {
         to->eps = from->eps;
-    if ((transfer & ECEF_SET) != 0) {
+    }
+    if (0 != (transfer & ECEF_SET)) {
         to->ecef.x = from->ecef.x;
         to->ecef.y = from->ecef.y;
         to->ecef.z = from->ecef.z;
         to->ecef.pAcc = from->ecef.pAcc;
     }
-    if ((transfer & VECEF_SET) != 0) {
+    if (0 != (transfer & VECEF_SET)) {
         to->ecef.vx = from->ecef.vx;
         to->ecef.vy = from->ecef.vy;
         to->ecef.vz = from->ecef.vz;
@@ -452,7 +598,7 @@ void gps_merge_fix(struct gps_fix_t *to,
             to->NED.relPosL = from->NED.relPosL;
         }
     }
-    if ((transfer & VNED_SET) != 0) {
+    if (0 != (transfer & VNED_SET)) {
         to->NED.velN = from->NED.velN;
         to->NED.velE = from->NED.velE;
         to->NED.velD = from->NED.velD;
@@ -462,11 +608,17 @@ void gps_merge_fix(struct gps_fix_t *to,
     }
     if (0 != isfinite(from->dgps_age) &&
         0 <= from->dgps_station) {
-        /* both, or neither */
+        // both, or neither
         to->dgps_age = from->dgps_age;
         to->dgps_station = from->dgps_station;
     }
 
+    if (ANT_UNK != from->ant_stat) {
+        to->ant_stat = from->ant_stat;
+    }
+    if (0 < from->jam) {
+        to->jam = from->jam;
+    }
     // navdata stuff.  just wind angle and angle for now
     if (0 != (transfer & NAVDATA_SET)) {
         if (0 != isfinite(from->wanglem)) {
@@ -485,6 +637,12 @@ void gps_merge_fix(struct gps_fix_t *to,
             to->wspeedt = from->wspeedt;
         }
     }
+    if (0 != isfinite(from->temp)) {
+        to->temp = from->temp;
+    }
+    if (0 != isfinite(from->wtemp)) {
+        to->wtemp = from->wtemp;
+    }
 }
 
 /* mkgmtime(tm)
@@ -493,6 +651,8 @@ void gps_merge_fix(struct gps_fix_t *to,
  * mktime() takes struct tm as localtime.
  *
  * The inverse of gmtime(time_t)
+ *
+ * Return: -1 on error, set errno to EOVERFLOW
  */
 time_t mkgmtime(struct tm * t)
 {
@@ -501,14 +661,37 @@ time_t mkgmtime(struct tm * t)
     static const int cumdays[MONTHSPERYEAR] =
         { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };
 
+    // check ranges, ignore tm_isdst and max tm_year
+    if (0 > t->tm_sec ||
+        0 > t->tm_min ||
+        0 > t->tm_hour ||
+        1 > t->tm_mday ||
+        0 > t->tm_mon ||
+        0 > t->tm_year ||
+        0 > t->tm_wday ||
+        0 > t->tm_yday ||
+        61 < t->tm_sec ||
+        59 < t->tm_min ||
+        23 < t->tm_hour ||
+        31 < t->tm_mday ||
+        11 < t->tm_mon ||
+        6 < t->tm_wday ||
+        365 < t->tm_yday) {
+        errno = EOVERFLOW;
+        return -1;
+    }
+    errno = 0;
     year = 1900 + t->tm_year + t->tm_mon / MONTHSPERYEAR;
     result = (year - 1970) * 365 + cumdays[t->tm_mon % MONTHSPERYEAR];
     result += (year - 1968) / 4;
     result -= (year - 1900) / 100;
     result += (year - 1600) / 400;
-    if ((year % 4) == 0 && ((year % 100) != 0 || (year % 400) == 0) &&
-        (t->tm_mon % MONTHSPERYEAR) < 2)
+    if (0 == (year % 4) &&
+        (0 != (year % 100) ||
+         0 == (year % 400)) &&
+        (2 > (t->tm_mon % MONTHSPERYEAR))) {
         result--;
+    }
     result += t->tm_mday - 1;
     result *= 24;
     result += t->tm_hour;
@@ -520,116 +703,39 @@ time_t mkgmtime(struct tm * t)
      * if (t->tm_isdst == 1)
      * result -= 3600;
      */
-    return (result);
+    return result;
 }
 
-timespec_t iso8601_to_timespec(char *isotime)
-/* ISO8601 UTC to Unix timespec, no leapsecond correction. */
+// ISO8601 UTC to Unix timespec, no leapsecond correction.
+timespec_t iso8601_to_timespec(const char *isotime)
 {
     timespec_t ret;
 
 #ifndef __clang_analyzer__
-#ifndef USE_QT
-    char *dp = NULL;
+#ifdef USE_QT
     double usec = 0;
-    struct tm tm;
-    memset(&tm,0,sizeof(tm));
 
-#ifdef HAVE_STRPTIME
-    dp = strptime(isotime, "%Y-%m-%dT%H:%M:%S", &tm);
-#else
-    /* Fallback for systems without strptime (i.e. Windows)
-     * This is a simplistic conversion for iso8601 strings only,
-     * rather than embedding a full copy of strptime() that handles
-     * all formats */
-    /* Thus avoiding needing to test for (broken) negative date/time
-     * numbers in token reading - only need to check the upper range */
-    bool failed = false;
-    char *isotime_tokenizer = strdup(isotime);
-    if (isotime_tokenizer) {
-      char *tmpbuf;
-      char *pch = strtok_r(isotime_tokenizer, "-T:", &tmpbuf);
-      int token_number = 0;
-
-      while (pch != NULL) {
-        double sec;
-        unsigned int tmp;
-
-        token_number++;
-        // Give up if encountered way too many tokens.
-        if (token_number > 10) {
-          failed = true;
-          break;
-        }
-        switch (token_number) {
-        case 1: // Year token
-          tmp = atoi(pch);
-          if (tmp < 9999)
-            tm.tm_year = tmp - 1900; // Adjust to tm year
-          else
-            failed = true;
-          break;
-        case 2: // Month token
-          tmp = atoi(pch);
-          if (tmp < 13)
-            tm.tm_mon = tmp - 1; // Month indexing starts from zero
-          else
-            failed = true;
-          break;
-        case 3: // Day token
-          tmp = atoi(pch);
-          if (tmp < 32)
-            tm.tm_mday = tmp;
-          else
-            failed = true;
-          break;
-        case 4: // Hour token
-          tmp = atoi(pch);
-          if (tmp < 24)
-            tm.tm_hour = tmp;
-          else
-            failed = true;
-          break;
-        case 5: // Minute token
-          tmp = atoi(pch);
-          if (tmp < 60)
-            tm.tm_min = tmp;
-          else
-            failed = true;
-          break;
-        case 6: // Seconds token
-          sec = safe_atof(pch);
-          // NB To handle timestamps with leap seconds
-          if (0 == isfinite(sec) &&
-              sec >= 0.0 && sec < 61.5 ) {
-            tm.tm_sec = (unsigned int)sec; // Truncate to get integer value
-            usec = sec - (unsigned int)sec; // Get the fractional part (if any)
-          }
-          else
-            failed = true;
-          break;
-        default: break;
-        }
-        pch = strtok_r(NULL, "-T:", &tmpbuf);
-      }
-      free(isotime_tokenizer);
-      // Split may result in more than 6 tokens if the TZ has any t's in it
-      // So check that we've seen enough tokens rather than an exact number
-      if (token_number < 6)
-        failed = true;
+    QString t(isotime);
+    QDateTime d = QDateTime::fromString(isotime, Qt::ISODate);
+    QStringList sl = t.split(".");
+    if (1 < sl.size()) {
+        usec = sl[1].toInt() / pow(10., (double)sl[1].size());
     }
-    if (failed)
-      memset(&tm,0,sizeof(tm));
-    else {
-      // When successful this normalizes tm so that tm_yday is set
-      //  and thus tm is valid for use with other functions
-      if (mktime(&tm) == (time_t)-1)
-        // Failed mktime - so reset the timestamp
-        memset(&tm,0,sizeof(tm));
+    ret.tv_sec = d.toTime_t();
+    ret.tv_nsec = usec * 1e9;
+#else  // USE_QT
+    double usec = 0;
+    struct tm tm = {0};
+
+    {
+        char *dp = NULL;
+        dp = strptime(isotime, "%Y-%m-%dT%H:%M:%S", &tm);
+        if (NULL != dp &&
+            '.' == *dp) {
+            usec = strtod(dp, NULL);
+        }
     }
-#endif
-    if (dp != NULL && *dp == '.')
-        usec = strtod(dp, NULL);
+
     /*
      * It would be nice if we could say mktime(&tm) - timezone + usec instead,
      * but timezone is not available at all on some BSDs. Besides, when working
@@ -638,24 +744,23 @@ timespec_t iso8601_to_timespec(char *isotime)
      * timezone database to get it right.
      */
     ret.tv_sec = mkgmtime(&tm);
-    ret.tv_nsec = usec * 1e9;;
-#else
-    double usec = 0;
+    ret.tv_nsec = usec * 1e9;
+#endif  // USE_QT
+#endif  // __clang_analyzer__
 
-    QString t(isotime);
-    QDateTime d = QDateTime::fromString(isotime, Qt::ISODate);
-    QStringList sl = t.split(".");
-    if (sl.size() > 1)
-        usec = sl[1].toInt() / pow(10., (double)sl[1].size());
-    ret.tv_sec = d.toTime_t();
-    ret.tv_nsec = usec * 1e9;;
+#if 4 < SIZEOF_TIME_T
+    if (253402300799LL < ret.tv_sec) {
+        // enforce max "9999-12-31T23:59:59.999Z"
+        ret.tv_sec = 253402300799LL;
+    }
 #endif
-#endif /* __clang_analyzer__ */
     return ret;
 }
 
-/* Unix timespec UTC time to ISO8601, no timezone adjustment */
-/* example: 2007-12-11T23:38:51.033Z */
+/* Convert POSIX timespec to ISO8601 UTC, put result in isotime.
+ * no timezone adjustment
+ * Return: pointer to isotime.
+ * example: 2007-12-11T23:38:51.033Z */
 char *timespec_to_iso8601(timespec_t fixtime, char isotime[], size_t len)
 {
     struct tm when;
@@ -664,20 +769,24 @@ char *timespec_to_iso8601(timespec_t fixtime, char isotime[], size_t len)
 
     if (0 > fixtime.tv_sec) {
         // Allow 0 for testing of 1970-01-01T00:00:00.000Z
-        return strncpy(isotime, "NaN", len);
+        strlcpy(isotime, "NaN", len);
+        return isotime;
     }
     if (999499999 < fixtime.tv_nsec) {
-        /* round up */
+        // round up
         fixtime.tv_sec++;
         fixtime.tv_nsec = 0;
     }
-#ifdef HAVE_GMTIME_R
-    (void)gmtime_r(&fixtime.tv_sec, &when);
-#else
-    /* Fallback to try with gmtime_s - primarily for Windows */
-    (void)gmtime_s(&when, &fixtime.tv_sec);
+
+#if 4 < SIZEOF_TIME_T
+    if (253402300799LL < fixtime.tv_sec) {
+        // enforce max "9999-12-31T23:59:59.999Z"
+        fixtime.tv_sec = 253402300799LL;
+    }
 #endif
 
+    (void)gmtime_r(&fixtime.tv_sec, &when);
+
     /*
      * Do not mess casually with the number of decimal digits in the
      * format!  Most GPSes report over serial links at 0.01s or 0.001s
@@ -691,8 +800,8 @@ char *timespec_to_iso8601(timespec_t fixtime, char isotime[], size_t len)
     return isotime;
 }
 
-/* return time now as ISO8601, no timezone adjustment */
-/* example: 2007-12-11T23:38:51.033Z */
+/* return time now as ISO8601, no timezone adjustment
+ * example: 2007-12-11T23:38:51.033Z */
 char *now_to_iso8601(char *tbuf, size_t tbuf_sz)
 {
     timespec_t ts_now;
@@ -741,8 +850,9 @@ double earth_distance_and_bearings(double lat1, double lon1,
                    (c_U1 * s_U2 - s_U1 * c_U2 * c_L) *
                    (c_U1 * s_U2 - s_U1 * c_U2 * c_L));
 
-        if (s_S == 0)
+        if (0 == s_S) {
             return 0;
+        }
 
         c_S = s_U1 * s_U2 + c_U1 * c_U2 * c_L;
         S = atan2(s_S, c_S);
@@ -750,17 +860,20 @@ double earth_distance_and_bearings(double lat1, double lon1,
         c_SqA = 1 - s_A * s_A;
         c_2SM = c_S - 2 * s_U1 * s_U2 / c_SqA;
 
-        if (0 == isfinite(c_2SM))
+        if (0 == isfinite(c_2SM)) {
             c_2SM = 0;
+        }
 
         C = f / 16 * c_SqA * (4 + f * (4 - 3 * c_SqA));
         L_P = lambda;
         lambda = L + (1 - C) * f * s_A *
             (S + C * s_S * (c_2SM + C * c_S * (2 * c_2SM * c_2SM - 1)));
-    } while ((fabs(lambda - L_P) > 1.0e-12) && (--i > 0));
+    } while ((fabs(lambda - L_P) > 1.0e-12) &&
+             (0 < --i));
 
-    if (i == 0)
+    if (0 == i) {
         return NAN;             // formula failed to converge
+    }
 
     uSq = c_SqA * ((a * a) - (b * b)) / (b * b);
     A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)));
@@ -769,21 +882,28 @@ double earth_distance_and_bearings(double lat1, double lon1,
                      (c_S * (-1 + 2 * c_2SM * c_2SM) - B / 6 * c_2SM *
                       (-3 + 4 * s_S * s_S) * (-3 + 4 * c_2SM * c_2SM)));
 
-    if (ib != NULL)
-        *ib = atan2(c_U2 * sin(lambda), c_U1 * s_U2 - s_U1 * c_U2 * cos(lambda));
-    if (fb != NULL)
-        *fb = atan2(c_U1 * sin(lambda), c_U1 * s_U2 * cos(lambda) - s_U1 * c_U2);
+    if (NULL != ib) {
+        *ib = atan2(c_U2 * sin(lambda),
+                    c_U1 * s_U2 - s_U1 * c_U2 * cos(lambda));
+    }
+    if (NULL != fb) {
+        *fb = atan2(c_U1 * sin(lambda),
+                    c_U1 * s_U2 * cos(lambda) - s_U1 * c_U2);
+    }
 
     return (WGS84B * A * (S - d_S));
 }
 
-/* Distance in meters between two points specified in degrees. */
+// Distance in meters between two points specified in degrees.
 double earth_distance(double lat1, double lon1, double lat2, double lon2)
 {
-        return earth_distance_and_bearings(lat1, lon1, lat2, lon2, NULL, NULL);
+    return earth_distance_and_bearings(lat1, lon1, lat2, lon2, NULL, NULL);
 }
 
-// Wait for data until timeout, ignoring signals.
+/* Wait for data until timeout, ignoring signals.
+ *
+ * pselect() may set errno on error
+ */
 bool nanowait(int fd, struct timespec *to)
 {
     fd_set fdset;
@@ -791,6 +911,7 @@ bool nanowait(int fd, struct timespec *to)
     FD_ZERO(&fdset);
     FD_SET(fd, &fdset);
     TS_NORM(to);         // just in case
+    errno = 0;
     // sigmask is NULL, so equivalent to select()
     return pselect(fd + 1, &fdset, NULL, NULL, to, NULL) == 1;
 }
@@ -835,11 +956,137 @@ void datum_code_string(int code, char *buffer, size_t len)
     }
 
     if (NULL == datum_str) {
-        /* Fake it */
+        // Fake it
         snprintf(buffer, len, "%d", code);
     } else {
         strlcpy(buffer, datum_str, len);
     }
 }
-/* end */
+
+/* make up an NMEA 4.0 (extended) PRN based on gnssId:svId,
+ * This does NOT match NMEA 4.10 and 4.11 where all PRN are 1-99,
+ * except IMEA
+ * Ref Appendix A from u-blox ZED-F9P Interface Description
+ * and
+ * Section 1.5.3  M10-FW500_InterfaceDescription_UBX-20053845.pdf
+ *
+ * Return PRN, or zero for error
+ */
+short ubx2_to_prn(int gnssId, int svId)
+{
+    short nmea_PRN;
+
+    if (1 > svId) {
+        // skip 0 svId
+        return 0;
+    }
+
+    switch (gnssId) {
+    case 0:
+        // GPS, 1-32 maps to 1-32
+        if (32 < svId) {
+            // skip bad svId
+            return 0;
+        }
+        nmea_PRN = svId;
+        break;
+    case 1:
+        // SBAS, 120..151, 152..158 maps to 33..64, 152..158
+        if (120 > svId) {
+            // Huh?
+            return 0;
+        } else if (151 >= svId) {
+            nmea_PRN = svId - 87;
+        } else if (158 >= svId) {
+            nmea_PRN = svId;
+        } else {
+            // Huh?
+            return 0;
+        }
+        break;
+    case 2:
+        // Galileo, ubx gnssid:svid 1..36 ->  301-336
+        // Galileo, ubx PRN         211..246 ->  301-336
+        if (36 >= svId) {
+            nmea_PRN = svId + 300;
+        } else if (211 > svId) {
+            // skip bad svId
+            return 0;
+        } else if (246 >= svId) {
+            nmea_PRN = svId + 90;
+        } else {
+            // skip bad svId
+            return 0;
+        }
+        break;
+    case 3:
+        // BeiDou, ubx gnssid:svid 1..37 -> to 401-437
+        // BeiDou, ubx PRN         159..163,33..64 -> to 401-437 ??
+        if (37 >= svId) {
+            nmea_PRN = svId + 400;
+        } else if (159 > svId) {
+            // skip bad svId
+            return 0;
+        } else if (163 >= svId) {
+            nmea_PRN = svId + 242;
+        } else {
+            // skip bad svId
+            return 0;
+        }
+        break;
+    case 4:
+        // IMES, ubx gnssid:svid 1-10 -> to 173-182
+        // IMES, ubx PRN         173-182 to 173-182
+        if (10 >= svId) {
+            nmea_PRN = svId + 172;
+        } else if (173 > svId) {
+            // skip bad svId
+            return 0;
+        } else if (182 >= svId) {
+            nmea_PRN = svId;
+        } else {
+            // > 182, skip bad svId
+            return 0;
+        }
+        break;
+    case 5:
+        // QZSS, ubx gnssid:svid 1-10 to 193-202
+        // QZSS, ubx PRN         193-202 to 193-202
+        if (10 >= svId) {
+            nmea_PRN = svId + 192;
+        } else if (193 > svId) {
+            // skip bad svId
+            return 0;
+        } else if (202 >= svId) {
+            nmea_PRN = svId;
+        } else {
+            // skip bad svId
+            return 0;
+        }
+        break;
+    case 6:
+        // GLONASS, 1-32 maps to 65-96
+        if (32 >= svId) {
+            nmea_PRN = svId + 64;
+        } else if (65 > svId) {
+            // skip bad svId
+            return 0;
+        } else if (96 >= svId) {
+            nmea_PRN = svId;
+        } else {
+            // skip bad svId, 255 == tracked, but unidentified, skip
+            return 0;
+        }
+        break;
+    case 7:
+        // NavIC (IRNSS)
+        FALLTHROUGH
+    default:
+        // Huh?
+        return 0;
+    }
+
+    return nmea_PRN;
+}
+
 // vim: set expandtab shiftwidth=4
